/****************************************************************

file swd_case2.c

Example Code searching for solutions of Sir Peter Swinnerton-Dyer's
Diophantine equation. An implementation of the hashing algorithm
developed by the authors.

Copyright (C) 2006 A.-S. Elsenhans and J. Jahnel

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

Address of the authors

	A.-S. Elsenhans/ J. Jahnel
	Math. Institut der Universitaet
	Bunsenstrasze 3--5
	D-37073 Goettingen
        Germany

WWW     http://www.uni-math.gwdg.de/jahnel

*****************************************************************/


#include"common.h"

#define HASH_GRENZE  ((1U << HASH_BREITE) - 1)
#define PUFFER_ZAHL  (1U << PUFFER_BITS)
#define PUFFER_LEN   (1U << (HASH_BREITE - PUFFER_BITS))
#define LEN           (1U << HASH_BREITE)

/* The hash table. */
signed int                                 hash_c[LEN];

/* Buffers. There are PUFFER_ZAHL of them.
   The same buffers are used for reading and writing.
   Without "+173" a lot slower in out tests?!
   Other numbers instead of 173: Often almost as fast. */
unsigned long int  puffer[PUFFER_ZAHL][PUFFER_LEN+173];
/* fenster_ende[i] always points to the first unoccupied entry in
   the buffer puffer[i]. */
unsigned long int           *fenster_ende[PUFFER_ZAHL];

/* Table of the fourth powers modulo PAGE_PRIME2. */
unsigned int                     h4_mod_p[PAGE_PRIME2];

/* Table of the fourth roots modulo PAGE_PRIME2. */
int                        vierte_wurzeln[PAGE_PRIME2];

/* Counters. */
long int                                           anz;
long int                                        h_used;

/* Constants for the Chinese Remainder Theorem.
   Modules: PAGE_PRIME2, 1250. */
long int                          eins_null, null_eins;
/* Modules: 10*PAGE_PRIME2, 81. */
long int                    eins_null_81, null_eins_81;

/* Number of clearings. Separate counters for each buffer.
   Necessary for the post processing. */
int                      puffer_leerungen[PUFFER_ZAHL];


/* The hash function. */
#define HASH(a)        (((unsigned int) (a)) >> (32 - HASH_BREITE))
/* Part of the hash function used for presorting. */
#define PUFFER_NR(a)   (((unsigned int) (a)) >> (32 - PUFFER_BITS))

/* CONTROL_FILTER = 2**n - 1 where n <= 31.
   For n = 32, the control value could become negative and -1 would
   lose the extra meaning "Place unoccupied". */
#define CONTROL_FILTER ((1U << 31) - 1)
/* The control function. */
#define CONTROL(a)     ((signed int) (((a) >> 32) & CONTROL_FILTER))


/* z = (a mod PAGE_PRIME2) and z = (b mod 1250). */
#define CHINESE_1250(a, b)                                        \
((((long int) a) * eins_null +                                    \
  ((long int) b) * null_eins) % (1250*PAGE_PRIME2))

/* z = (a mod 10*PAGE_PRIME2) and z = (b mod 81). */
#define CHINESE_81(a, b)                                          \
((((long int) a) * eins_null_81 +                                 \
  ((long int) b) * null_eins_81) % (810*PAGE_PRIME2))


/* Tests for PAGE_PRIME2.
   Tabulation of the fourth powers modulo PAGE_PRIME2.
   Initialization of the constants for the Chinese Remainder
   Theorem. */
void init () {
 long int  i, t;

 /* Testing PAGE_PRIME2. */
 if (1250*PAGE_PRIME2 <= SUCHWEITE) {
  printf ("PAGE_PRIME2 must be bigger than SUCHWEITE/1250.");
  exit (0);
 }
 if ((PAGE_PRIME2 % 4) != 3) {
  printf ("PAGE_PRIME2 must be 3 modulo 4.\n");
  exit (0);
 }
 for (i = 3; i*i <= PAGE_PRIME2; i += 2) {
  if (PAGE_PRIME2 % i == 0) {
   printf ("PAGE_PRIME2 must a prime number.\n");
   exit (0);
  }
 }

 /* Initialization of the counter for coincidences. */
 anz = 0;

 /* Tabulate the fourth powers modulo PAGE_PRIME2. */
 for (i = 0; i < PAGE_PRIME2; i++) {
  t = (i*i) % PAGE_PRIME2;
  t = (t*t) % PAGE_PRIME2;
  h4_mod_p[i] = t;           /* Fourth powers modulo PAGE_PRIME2. */
 }

 for (i = 0; i < PAGE_PRIME2; i++)
  vierte_wurzeln[i] = -1;

 for (i = 0; i < PAGE_PRIME2; i++)
  vierte_wurzeln[h4_mod_p[i]] = i;

 /* Initialize the constants for the Chinese Remainder Theorem. */
 eins_null = 1;
 while ((eins_null % 1250) != 0)
  eins_null += PAGE_PRIME2;
 null_eins = 1;
 while ((null_eins % PAGE_PRIME2) != 0)
  null_eins += 1250;

 eins_null_81 = 1;
 while ((eins_null_81 % 81) != 0)
  eins_null_81 += 10*PAGE_PRIME2;
 null_eins_81 = 1;
 while ((null_eins_81 % (10*PAGE_PRIME2)) != 0)
  null_eins_81 += 81;
}


/* CLEANING UP */

/* Declare the buffers to be empty.
   Initialization of puffer_leerungen for a new residue class. */
void puffer_init () {
 int  i;

 for (i = 0; i < PUFFER_ZAHL; i++) {
  fenster_ende[i] = &puffer[i][0];
  puffer_leerungen[i] = 0;
 }
}


/* Declare the hash table to be empty. */
void hashtabelle_init () {
 int  i;

 h_used = 0;

 for (i = 0; i < LEN; i++)
  hash_c[i] = -1;
}


/* WRITING */

/* Inclusion of a single pair into the buffer. */
inline void z_und_w_schreiben (int z, int w) {
 unsigned long int  z_lang, w_lang, rechte_seite;
 unsigned int                                  t;

 if (z <= SUCHWEITE) {
  z_lang = z;
  w_lang = w;

  rechte_seite =
   z_lang*z_lang*z_lang*z_lang + 4 * w_lang*w_lang*w_lang*w_lang;
  /* modulo 2**64. */

  t = PUFFER_NR (rechte_seite);

  /* This means an overflow during the writing step.
     Should never happen. */
  if (fenster_ende[t] >= &puffer[t][0] + PUFFER_LEN) {
   printf ("Overflow in one of the buffers!\n");
   exit (0);
  }

  /* Write rechte_seite into the buffer. */
  *(fenster_ende[t]) = rechte_seite;
  fenster_ende[t]++;

  h_used++;
 }
}


/* Build-up of the buffer for writing.
   We assume that 1250*PAGE_PRIME2 > SUCHWEITE. */
void schreibdaten_erzeugen_und_puffern (int rkl) {
 int       z0, z1, w, w0;
 int  vw1, vw2, vw3, vw4;

 for (w0 = 0; w0 < 1250; w0 += 2) {
  if ((w0 % 5) != 0) {
   vw1 = (625 + (181 * w0)) % 1250;
   vw2 = 1250 - vw1;
   vw3 = (625 + (183 * w0)) % 1250;
   vw4 = 1250 - vw3;
   /* We have z**4 = -4 mod 625 if and only if z = \pm181
      or z = \pm183. We add 625 since w is even and z belongs to an
      odd residue class modulo 1250. */

   for (w = w0; w <= SUCHWEITE; w += 1250) {
    z0 = (rkl + 4 * (PAGE_PRIME2 - h4_mod_p[w % PAGE_PRIME2]))
                                                    % PAGE_PRIME2;
    z0 = vierte_wurzeln[z0];

    /* If there exists a fourth root modulo PAGE_PRIME2. */
    if (z0 != -1) {
     z1 = CHINESE_1250 (z0, vw1);
     z_und_w_schreiben (z1, w);
     if (z1 != 0) {
      z1 = 1250*PAGE_PRIME2 - z1;
      z_und_w_schreiben (z1, w);
     }

     z1 = CHINESE_1250 (z0, vw2);
     z_und_w_schreiben (z1, w);
     if (z1 != 0) {
      z1 = 1250*PAGE_PRIME2 - z1;
      z_und_w_schreiben (z1, w);
     }

     z1 = CHINESE_1250 (z0, vw3);
     z_und_w_schreiben (z1, w);
     if (z1 != 0) {
      z1 = 1250*PAGE_PRIME2 - z1;
      z_und_w_schreiben (z1, w);
     }

     z1 = CHINESE_1250 (z0, vw4);
     z_und_w_schreiben (z1, w);
     if (z1 != 0) {
      z1 = 1250*PAGE_PRIME2 - z1;
      z_und_w_schreiben (z1, w);
     }
    } /* End-if to if (z0 != -1). */
   }  /* To for (w = w0; ... ). */
  }   /* End-if to if ((w0 % 5) != 0). */
 }    /* To for (w0 = 0; ... ). */
}


/* Creation of the hash table from the buffer. */
void hashtabelle_schreiben () {
 int                                i, h;
 unsigned long int  rechte_seite, *index;

 for (i = 0; i < PUFFER_ZAHL; i++) {
  for (index = &puffer[i][0]; index < fenster_ende[i]; index++) {
   rechte_seite = *index;         /* rechte_seite = puffer[i][k] */
   h = HASH (rechte_seite);

   /* Resolution of collisions. */
   while (hash_c[h] != -1)
    h = ((h + 1) & HASH_GRENZE);   /* (h+1) mod (2**HASH_BREITE) */

   /* Inclusion. */
   hash_c[h] = CONTROL (rechte_seite);
  }
 }
 /* Hash built. */
}


/* READING */

/* One step towards building up the buffers for reading.
   y is fixed, x is free in one residue class.
   Start the actual reading when a buffer fills up.  */
void restklasse_von_x_mit_y_fest_testen (int x, int y, int klasse) {
 unsigned long int  linke_seite, x_lang, y_lang, y4;
 unsigned int                            x_10, x_30;
 signed int                                       f;

 x_10 = x;
 while ((x_10 % 10) != 5)
  x_10 += PAGE_PRIME2;

 y_lang = y;
 y4 = 2 * (y_lang*y_lang) * (y_lang*y_lang);     /* modulo 2**64. */

 /* Reading is necessary
       if ((x % 3 == 0) || (y % 3 == 0)
                  || ((x+34*y) % 81 == 0) || ((x-34*y) % 81 == 0)).

    In the first two cases, 3 \not| (x**4 + 2 y**4) or
    81 | (x**4 + 2 y**4), anyway.

    On the other hand, if x and y are not divisible by 3 then,
    automatically, 3 | (x**4 + 2 y**4) while
       81 | (x**4 + 2 y**4) <==> x = \pm 34*y (mod 81).

    Observe: (Z/81Z)* is cyclic of order 54.
              Since 4 \not| 54, each element has at most two
              fourth roots.
              More concretely, \pm34**4 = -2 mod 81. */
 if (y % 3 == 0) {
  /* In this case, there is no condition modulo 3 oder 81. */
  for (x_lang = x_10; x_lang <= SUCHWEITE; x_lang += 10*PAGE_PRIME2)
  {
   linke_seite = (x_lang*x_lang) * (x_lang*x_lang) + y4;
   /* modulo 2**64. */

   f = PUFFER_NR (linke_seite);

   /* If the buffer is full. */
   if (fenster_ende[f] >= &puffer[f][0] + PUFFER_LEN) {
    einen_puffer_abarbeiten (f, klasse);
    /* This leaves us with an empty buffer. */
   }

   *(fenster_ende[f]) = linke_seite;
   fenster_ende[f]++;
  } /* End of the for-loop for x. */
 }
 else { /* Here, 3|x or x = \pm 34*y (mod 81). */
  /* The case 3|x. */
  x_30 = x_10;
  while (x_30 % 3 != 0)
   x_30 += 10*PAGE_PRIME2;
  for (x_lang = x_30; x_lang <= SUCHWEITE; x_lang += 30*PAGE_PRIME2)
  {
   linke_seite = (x_lang*x_lang) * (x_lang*x_lang) + y4;
   /* modulo 2**64. */
   f = PUFFER_NR (linke_seite);

   /* If the buffer is full. */
   if (fenster_ende[f] >= &puffer[f][0] + PUFFER_LEN) {
    einen_puffer_abarbeiten (f, klasse);
    /* This leaves us with an empty buffer. */
   }

   *(fenster_ende[f]) = linke_seite;
   fenster_ende[f]++;
  }  /* End of the for-loop for x. */

  /* The case x = 34*y (mod 81).
     Hence, 3 \not| y implies 3 \not| x. */
  x_lang = CHINESE_81 (x_10, 34*y_lang);
  for (; x_lang <= SUCHWEITE; x_lang += 810*PAGE_PRIME2) {
   /* This loop will not be run more than once. */
   linke_seite = (x_lang*x_lang) * (x_lang*x_lang) + y4;
   /* modulo 2**64. */
   f = PUFFER_NR (linke_seite);

   /* If the buffer is full. */
   if (fenster_ende[f] >= &puffer[f][0] + PUFFER_LEN) {
    einen_puffer_abarbeiten (f, klasse);
    /* This leaves us with an empty buffer. */
   }

   *(fenster_ende[f]) = linke_seite;
   fenster_ende[f]++;
  }  /* End of the for-loop for x. */

  /* The case x = 47*y = -34*y (mod 81).
     Hence, 3 \not| y implies 3 \not| x. */
  x_lang = CHINESE_81 (x_10, 47*y_lang);
  for (; x_lang <= SUCHWEITE; x_lang += 810*PAGE_PRIME2) {
   /* This loop will not be run more than once. */
   linke_seite = (x_lang*x_lang) * (x_lang*x_lang) + y4;
   /* modulo 2**64. */
   f = PUFFER_NR (linke_seite);

   /* If the buffer is full. */
   if (fenster_ende[f] >= &puffer[f][0] + PUFFER_LEN) {
    einen_puffer_abarbeiten (f, klasse);
    /* This leaves us with an empty buffer. */
   }

   *(fenster_ende[f]) = linke_seite;
   fenster_ende[f]++;
  }  /* End of the for-loop for x. */
 }
}


/* Build-up of the buffers for reading.
   The data in the buffers are actually read automatically every
   time when a buffer fills up. */
void lesedaten_erzeugen_und_lesen (int klasse) {
 int  y0, y, x0, t;

 for (x0 = 0; x0 < PAGE_PRIME2; x0++) {
  t = (PAGE_PRIME2 + klasse - h4_mod_p[x0 % PAGE_PRIME2]) % PAGE_PRIME2;
  if (t % 2 == 0)
   t = t / 2;
  else
   t = (t + PAGE_PRIME2) / 2;

  y0 = vierte_wurzeln[t];

  if (y0 != - 1) { /* If there exists a fourth root of t. */
   y = y0;
   while ((y % 10) != 0)
    y += PAGE_PRIME2;
   while (y <= SUCHWEITE) {
    restklasse_von_x_mit_y_fest_testen (x0, y, klasse);
    y += 10*PAGE_PRIME2;
   }

   if (y0 > 0) { /* If the fourth root of t is different from 0.
                    In this case, there are exactly two of them. */
    y = PAGE_PRIME2 - y0;
    while ((y % 10) != 0)
     y += PAGE_PRIME2;
    while (y <= SUCHWEITE) {
     restklasse_von_x_mit_y_fest_testen (x0, y, klasse);
     y += 10*PAGE_PRIME2;
    }
   }  /* End-if to if (y0 > 0). */
  }   /* End-if to if (y0 != -1). */
 }    /* To for (x0 = 0; ... ). */
}


/* All data in buffer number f_nr are looked up in the hash table.
   When a coincidence is detected, post processing is called.

   Afterwards, the buffer is declared to be empty.

   This function is called every time for a single buffer when it
   is full. At the very end, i.e. when the last reading data have
   been written into the buffer, it is called for all buffers. */
void einen_puffer_abarbeiten (int f_nr, int klasse) {
 unsigned long int  linke_seite, *index;
 signed int                           c;
 int                                  t;

 for (index = &puffer[f_nr][0]; index < fenster_ende[f_nr]; index++)
 {
  linke_seite = *index;          /* linke_seite = puffer[f_nr][k] */
  t = HASH (linke_seite);

  /* The actual reading. */
  /* Optimization for speed. Some sort of loop unrolling.
     If we hit an unoccupied place then that reading access finishes
     directly at this line. */
  if (hash_c[t] >= 0) {
   c = CONTROL (linke_seite);
   do {
    /* Found an interesting value. */
    if (hash_c[t] == c)
     post_proc (linke_seite, klasse, f_nr,
       PUFFER_LEN * puffer_leerungen[f_nr]
      + (index - &puffer[f_nr][0]));
     /* The last operand is the absolute buffer position which would
        have been reached without any clearings. */
    t = ((t + 1) & HASH_GRENZE);    /* (t+1) mod (2**HASH_BREITE) */
   } while (hash_c[t] != -1);     /* Observe possible collisions. */
  }  /* End-if to if (hash_c[t] >= 0). */
 }   /* To for (index = &puffer[f_nr][0]; ... ). */

 /* Declare the buffer to be empty. */
 fenster_ende[f_nr] = &puffer[f_nr][0];
 puffer_leerungen[f_nr]++;  /* Count the number of all clearings. */
}


/* Looking up all buffered data in the hash table.
   Clearing all buffers. */
void alle_puffer_abarbeiten (int klasse) {
 int  i;

 for (i = 0; i < PUFFER_ZAHL; i++)
  if (fenster_ende[i] > &puffer[i][0])
   einen_puffer_abarbeiten (i, klasse);
}


/* The outer loop. */
int main (int argc, char *argv[]) {
 int  rkl, start, ende;

 if (argc != 3) {
  printf ("Call needs 2 parameters.\n");
  exit (0);
 }
 start = atoi (argv[1]);
 ende = atoi (argv[2]);
 if (start < 0)
  start = 0;
 if (ende > (PAGE_PRIME2 - 1))
  ende = PAGE_PRIME2 - 1;
 if (start > ende) {
  printf ("Range from %d to %d is empty.\n", start, ende);
  exit (0);
 }

 printf ("Range from %d to %d.\n", start, ende);
 printf ("Initialize tables of powers and roots.\n");
 init ();
 printf ("Initialization finished.\n");

 for (rkl = start; rkl <= ende; rkl++) {
  /* Clearing up. */
  hashtabelle_init ();
  puffer_init ();

  /* Writing. */
  /* Build up the buffer in residue class rkl. */
  schreibdaten_erzeugen_und_puffern (rkl);
  /* Creating the hash table. */
  hashtabelle_schreiben ();
  printf ("Searching in residue class %d. ", rkl);
  printf ("%ld entries in the hash table.\n", h_used);

  /* Clearing up. */
  puffer_init ();

  /* Reading. */
  /* Build up the buffers in residue class rkl.
     The actual reading is done automatically in blocks
     corresponding to full buffers. */
  lesedaten_erzeugen_und_lesen (rkl);
  /* Clear all the buffers and read. */
  alle_puffer_abarbeiten (rkl);
  fflush (stdout);
 }

 printf ("Totally %ld coincidences detected.\n", anz);
 return 0;
}


/* POST PROCESSING */

/* The whole code below is only for post processing.
   It is not called too often. */

/* This function is one step of the simulated build up of the
   buffers for reading. y is fixed, x is free in one residue
   class. */
int restklasse_von_x_mit_y_fest_probew_testen
   (int x, int y, int klasse, int f_nr, int i, int *puffer_index,
    int *x_gefunden, int *y_gefunden) {
 unsigned long int  linke_seite, x_lang, y_lang, y4;
 unsigned int                            x_10, x_30;
 signed int                                       f;

 x_10 = x;
 while ((x_10 % 10) != 5)
  x_10 += PAGE_PRIME2;

 y_lang = y;
 y4 = 2 * (y_lang*y_lang) * (y_lang*y_lang);     /* modulo 2**64. */

 /* Reading is necessary
       if ((x % 3 == 0) || (y % 3 == 0)
                  || ((x+34*y) % 81 == 0) || ((x-34*y) % 81 == 0)).

    In the first two cases, 3 \not| (x**4 + 2 y**4) or
    81 | (x**4 + 2 y**4), anyway.

    On the other hand, if x and y are not divisible by 3 then,
    automatically, 3 | (x**4 + 2 y**4) while
       81 | (x**4 + 2 y**4) <==> x = \pm 34*y (mod 81). */
 if (y % 3 == 0) {
  /* In this case, there is no condition modulo 3 oder 81. */
  for (x_lang = x_10; x_lang <= SUCHWEITE; x_lang += 10*PAGE_PRIME2)
  {
   linke_seite = (x_lang*x_lang) * (x_lang*x_lang) + y4;
   /* modulo 2**64. */
   f = PUFFER_NR (linke_seite);
   if (f == f_nr) {
    (*puffer_index)++;
    if (*puffer_index == i) {
     *x_gefunden = x_lang;
     *y_gefunden = y;
     return 0;
    }
   }
  } /* End of the for-loop for x. */
 }
 else { /* Here, 3|x or x = \pm 34*y (mod 81). */
  /* The case 3|x. */
  x_30 = x_10;
  while (x_30 % 3 != 0)
   x_30 += 10*PAGE_PRIME2;
  for (x_lang = x_30; x_lang <= SUCHWEITE; x_lang += 30*PAGE_PRIME2)
  {
   linke_seite = (x_lang*x_lang) * (x_lang*x_lang) + y4;
   /* modulo 2**64. */
   f = PUFFER_NR (linke_seite);
   if (f == f_nr) {
    (*puffer_index)++;
    if (*puffer_index == i) {
     *x_gefunden = x_lang;
     *y_gefunden = y;
     return 0;
    }
   }
  } /* End of the for-loop for x. */

  /* The case x = 34*y (mod 81).
     Hence, 3 \not| y implies 3 \not| x. */
  x_lang = CHINESE_81 (x_10, 34*y_lang);
  for (; x_lang <= SUCHWEITE; x_lang += 810*PAGE_PRIME2) {
   /* This loop will not be run more than once. */
   linke_seite = (x_lang*x_lang) * (x_lang*x_lang) + y4;
   /* modulo 2**64. */
   f = PUFFER_NR (linke_seite);
   if (f == f_nr) {
    (*puffer_index)++;
    if (*puffer_index == i) {
     *x_gefunden = x_lang;
     *y_gefunden = y;
     return 0;
    }
   }
  }  /* End of the for-loop for x. */

  /* The case x = 47*y = -34*y (mod 81).
     Hence, 3 \not| y implies 3 \not| x. */
  x_lang = CHINESE_81 (x_10, 47*y_lang);
  for (; x_lang <= SUCHWEITE; x_lang += 810*PAGE_PRIME2) {
   /* This loop will not be run more than once. */
   linke_seite = (x_lang*x_lang) * (x_lang*x_lang) + y4;
   /* modulo 2**64. */
   f = PUFFER_NR (linke_seite);
   if (f == f_nr) {
    (*puffer_index)++;
    if (*puffer_index == i) {
     *x_gefunden = x_lang;
     *y_gefunden = y;
     return 0;
    }
   }
  }  /* End of the for-loop for x. */
 }

 /* Return value 1 means "Not yet found".
    We return 0 as soon as the correct absolute buffer position is
    reached again. This means, immediately when x and y are
    found. */
 return 1;
}


/* Simulation of the build-up of the buffer for reading.

   This function is similar to lesedaten_erzeugen_und_lesen.
   It does, however, never write into memory. It only counts how
   often an entry would be written into buffer number f_nr.
   It finds the pair (w, y) which generates the i-th entry into
   buffer number f_nr. */
int lesepufferaufbau_simulieren (int klasse, int f_nr, int i,
    int *x_gefunden, int *y_gefunden) {
 int  y0, y, x0, t, puffer_index;

 puffer_index = -1;

 for (x0 = 0; x0 < PAGE_PRIME2; x0++) {
  t = (PAGE_PRIME2 + klasse - h4_mod_p[x0 % PAGE_PRIME2]) % PAGE_PRIME2;
  if (t % 2 == 0)
   t = t / 2;
  else
   t = (t + PAGE_PRIME2) / 2;

  y0 = vierte_wurzeln[t];

  if (y0 != - 1) { /* If there exists a fourth root of t. */
   y = y0;
   while ((y % 10) != 0)
    y += PAGE_PRIME2;
   while (y <= SUCHWEITE) {
    /* Return value 1 means "found". */
    if (restklasse_von_x_mit_y_fest_probew_testen
        (x0, y, klasse, f_nr, i, &puffer_index,
         x_gefunden, y_gefunden) == 0)
     return 0;
    y += 10*PAGE_PRIME2;
   }

   if (y0 > 0) { /* If the fourth root of t is different from 0.
                    In this case, there are exactly two of them. */
    y = PAGE_PRIME2 - y0;
    while ((y % 10) != 0)
     y += PAGE_PRIME2;
    while (y <= SUCHWEITE) {
     /* Return value 1 means "found". */
     if (restklasse_von_x_mit_y_fest_probew_testen
         (x0, y, klasse, f_nr, i, &puffer_index,
          x_gefunden, y_gefunden) == 0)
      return 0;
     y += 10*PAGE_PRIME2;
    }
   }  /* End-if to if (y0 > 0). */
  }   /* End-if to if (y0 != -1). */
 }    /* To for x0. */

 /* This line should never be reached. When (x, y) is reconstructed
    then return 0 is carried out immediately. The return value 1
    allows easily to detect an error. */
 return 1;
}


/* This function calculates for a pair (z, w) whether the control
   value is equal to c. If this happens then an output is
   generated. */
inline void z_und_w_testen (int x, int y, int z, int w, signed int c) {
 unsigned long int x_lang, y_lang, z_lang, w_lang;
 unsigned long int           argh, argl, lsh, lsl;
 signed int                                c_wert;

 x_lang = x;
 y_lang = y;
 z_lang = z;
 w_lang = w;

 if (z <= SUCHWEITE) {
  umul_ppmm (lsh, lsl, z_lang * z_lang, z_lang * z_lang);
  umul_ppmm (argh, argl, (w_lang * w_lang) << 1,
                         (w_lang * w_lang) << 1);
  add_ssaaaa (lsh, lsl, lsh, lsl, argh, argl);
  /* (lsh, lsl) = z**4 + 4 w**4. */

  c_wert = CONTROL (lsl);

  if (c == c_wert) {
   printf ("x = %d, y = %d, z = %d, w = %d.\n", x, y, z, w);

   umul_ppmm (argh, argl, x_lang * x_lang, x_lang * x_lang);
   sub_ddmmss (lsh, lsl, lsh, lsl, argh, argl);
   umul_ppmm (argh, argl, y_lang * y_lang, y_lang * y_lang);
   add_ssaaaa (argh, argl, argh, argl, argh, argl);
   sub_ddmmss (lsh, lsl, lsh, lsl, argh, argl);
   /* (lsh, lsl) = z**4 + 4 w**4 - x**4 - 2 y**4. */

   add_ssaaaa (lsh, lsl, lsh, lsl, lsh, lsl);
   /* This does not cause an overflow as long as
                SUCHWEITE < [2**128/10]**(1/4) = 2.41523760e9. */
   printf ("The hash deviation is %ld.\n", ((long int) lsl) / 2);
   printf ("The deviation in size is %ld * 2**63.\n", lsh);
   anz++;
  }
 }
}


/* Search for z and w such that the control value fits with ls. */
long int z_und_w_suchen (unsigned long int ls, int klasse,
         int x_gefunden, int y_gefunden) {
 long int              anz_alt;
 int             z0, z1, w, w0;
 int        vw1, vw2, vw3, vw4;
 signed int                  c;

 anz_alt = anz;
 c = CONTROL (ls);

 for (w0 = 0; w0 < 1250; w0 += 2) {
  if ((w0 % 5) != 0) {
   vw1 = (625 + (181 * w0)) % 1250;
   vw2 = 1250 - vw1;
   vw3 = (625 + (183 * w0)) % 1250;
   vw4 = 1250 - vw3;

   for (w = w0; w <= SUCHWEITE; w += 1250) {
    z0 = (klasse + 4 * (PAGE_PRIME2 - h4_mod_p[w % PAGE_PRIME2]))
                                                      % PAGE_PRIME2;
    z0 = vierte_wurzeln[z0];

    /* If there exists a fourth root modulo PAGE_PRIME2. */
    if (z0 != -1) {
     z1 = CHINESE_1250 (z0, vw1);
     z_und_w_testen (x_gefunden, y_gefunden, z1, w, c);
     if (z1 != 0) {
      z1 = 1250*PAGE_PRIME2 - z1;
      z_und_w_testen (x_gefunden, y_gefunden, z1, w, c);
     }

     z1 = CHINESE_1250 (z0, vw2);
     z_und_w_testen (x_gefunden, y_gefunden, z1, w, c);
     if (z1 != 0) {
      z1 = 1250*PAGE_PRIME2 - z1;
      z_und_w_testen (x_gefunden, y_gefunden, z1, w, c);
     }

     z1 = CHINESE_1250 (z0, vw3);
     z_und_w_testen (x_gefunden, y_gefunden, z1, w, c);
     if (z1 != 0) {
      z1 = 1250*PAGE_PRIME2 - z1;
      z_und_w_testen (x_gefunden, y_gefunden, z1, w, c);
     }

     z1 = CHINESE_1250 (z0, vw4);
     z_und_w_testen (x_gefunden, y_gefunden, z1, w, c);
     if (z1 != 0) {
      z1 = 1250*PAGE_PRIME2 - z1;
      z_und_w_testen (x_gefunden, y_gefunden, z1, w, c);
     }
    } /* End-if to if (z0 != -1). */
   }  /* To for (w = w0; ... ). */
  }   /* End-if to if ((w0 % 5) != 0). */
 }    /* To for (w0 = 0; ... ). */

 /* The return value is the number of the pairs (z, w) fitting. */
 return (anz - anz_alt);
}


/* Post processing.

   Expects linke_seite in ls, (ls mod PAGE_PRIME2) in klasse, the
   number of the window in f_nr, and the absolute position of the
   entry causing the coincidence in i. */
void post_proc (unsigned long int ls, int klasse, int f_nr, int i) {
 int  x_gef, y_gef;

 /* Coincidence in residue class klasse, linke_seite = ls. */

 /* Reconstruction of x and y.
    Exiting with an error message if (x, y) cannot be
    reconstructed. */
 if (lesepufferaufbau_simulieren (klasse, f_nr, i, &x_gef, &y_gef))
 {
  printf ("Error in the buffer for reading. Exiting.\n");
  exit (0);
 }

 /* Reconstruction of z und w. It is possible that more than one
    pair is found. */
 if (z_und_w_suchen (ls, klasse, x_gef, y_gef) == 0) {
  /* Exiting with an error message if no pair (z, w) is found. */
  printf ("Error! Exiting.\n");
  exit (0);
 }
}
